#!/usr/bin/env python3

#    Kosmorro - Compute The Next Ephemerides
#    Copyright (C) 2019  Jérôme Deuchnord <jerome@deuchnord.fr>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import argcomplete
import sys
import os.path

import pytz
from babel.dates import format_date
from kosmorrolib import Position, get_ephemerides, get_events, get_moon_phase
from kosmorrolib.exceptions import OutOfRangeDateError
from datetime import date

from . import dumper, environment, debug
from .date import parse_date
from .geolocation import get_position
from .utils import (
    KOSMORRO_VERSION,
    KOSMORROLIB_VERSION,
    colored,
    set_colors_activated,
    print_stderr,
    get_timezone,
)
from .exceptions import (
    InvalidOutputFormatError,
    UnavailableFeatureError,
    OutOfRangeDateError as DateRangeError,
)
from kosmorro.i18n.utils import _


def run():
    env_vars = environment.get_env_vars()
    output_formats = get_dumpers()
    args = get_args(list(output_formats.keys()))
    debug.show_debug_messages = args.show_debug_messages
    output_format = args.format

    set_colors_activated(args.colors)

    if args.completion is not None:
        return 0 if output_completion(args.completion) else 1

    if args.special_action is not None:
        return 0 if args.special_action() else 1

    try:
        compute_date = parse_date(args.date)
    except ValueError as error:
        print_stderr(colored(error.args[0], color="red", attrs=["bold"]))
        return -1

    position = None
    try:
        if args.position not in [None, ""]:
            position = get_position(args.position)
        elif env_vars.position not in [None, ""]:
            position = get_position(env_vars.position)
    except ValueError as e:
        print_stderr(colored(e.args[0], color="red", attrs=["bold"]))

    # if output format is not specified, try to use output file extension as output format
    if args.output is not None and output_format is None:
        file_extension = os.path.splitext(args.output)[-1][1:].lower()
        if file_extension:
            output_format = file_extension

    # default to .txt if output format was not given and output file did not have file extension
    if output_format is None:
        output_format = "txt"

    if output_format == "pdf":
        print(
            _(
                "Save the planet and paper!\n"
                "Consider printing your PDF document only if really necessary, and use the other side of the sheet."
            )
        )
        if position is None:
            print_stderr(
                colored(
                    _(
                        "PDF output will not contain the ephemerides, because you didn't provide the observation "
                        "coordinates."
                    ),
                    "yellow",
                )
            )

    timezone = 0

    try:
        if args.timezone is not None:
            timezone = get_timezone(args.timezone, compute_date)
        elif env_vars.tz is not None:
            timezone = get_timezone(env_vars.tz, compute_date)
        elif env_vars.timezone is not None:
            print_stderr(
                colored(
                    _(
                        "Environment variable KOSMORRO_TIMEZONE is deprecated. Use TZ instead, which is more standard."
                    ),
                    "yellow",
                )
            )
            timezone = get_timezone(env_vars.timezone, compute_date)
    except pytz.UnknownTimeZoneError as error:
        print_stderr(
            colored(
                _("Unknown timezone: {timezone}").format(timezone=error.args[0]),
                color="red",
            )
        )
        return -1

    try:
        use_colors = not environment.NO_COLOR and args.colors

        output = get_information(
            compute_date,
            position,
            timezone,
            output_format,
            use_colors,
            args.show_graph,
        )
    except InvalidOutputFormatError as error:
        print_stderr(colored(error.msg, "red"))
        debug.debug_print(error)
        return 3
    except UnavailableFeatureError as error:
        print_stderr(colored(error.msg, "red"))
        debug.debug_print(error)
        return 2
    except DateRangeError as error:
        print_stderr(colored(error.msg, "red"))
        debug.debug_print(error)
        return 1

    if args.output is not None:
        try:
            file_content = output.to_string()
            opening_mode = get_opening_mode(output_format)
            with open(args.output, opening_mode) as output_file:
                output_file.write(file_content)
        except UnavailableFeatureError as error:
            print_stderr(colored(error.msg, "red"))
            debug.debug_print(error)
            return 2
        except OSError as error:
            print_stderr(
                colored(
                    _('The file could not be saved in "{path}": {error}').format(
                        path=args.output, error=error.strerror
                    ),
                    "red",
                )
            )
            debug.debug_print(error)

            return 3
    elif not output.is_file_output_needed():
        print(output)
    else:
        print_stderr(
            colored(
                _("Please provide a file path to export in this format (--output)."),
                color="red",
            )
        )
        return 1

    return 0


def get_information(
    compute_date: date,
    position: Position,
    timezone: int,
    output_format: str,
    colors: bool,
    show_graph: bool,
) -> dumper.Dumper:
    try:
        if position is not None:
            eph = get_ephemerides(
                for_date=compute_date, position=position, timezone=timezone
            )
        else:
            eph = []

        try:
            moon_phase = get_moon_phase(for_date=compute_date, timezone=timezone)
        except OutOfRangeDateError as error:
            moon_phase = None
            print_stderr(
                colored(
                    _(
                        "Moon phase can only be computed between {min_date} and {max_date}"
                    ).format(
                        min_date=format_date(error.min_date, "long"),
                        max_date=format_date(error.max_date, "long"),
                    ),
                    "yellow",
                )
            )

        events_list = get_events(compute_date, timezone)

        return get_dumpers()[output_format](
            ephemerides=eph,
            moon_phase=moon_phase,
            events=events_list,
            date=compute_date,
            timezone=timezone,
            with_colors=colors,
            show_graph=show_graph,
        )
    except KeyError as error:
        raise InvalidOutputFormatError(output_format, list(get_dumpers().keys()))
    except OutOfRangeDateError as error:
        raise DateRangeError(error.min_date, error.max_date)


def get_dumpers() -> {str: dumper.Dumper}:
    return {
        "txt": dumper.TextDumper,
        "json": dumper.JsonDumper,
        "pdf": dumper.PdfDumper,
        "tex": dumper.LatexDumper,
    }


def get_opening_mode(format: str) -> str:
    if format == "pdf":
        return "wb"

    return "w"


def output_version() -> bool:
    python_version = "%d.%d.%d" % (
        sys.version_info[0],
        sys.version_info[1],
        sys.version_info[2],
    )
    print("Kosmorro %s" % KOSMORRO_VERSION)
    print(
        _(
            "Running on Python {python_version} "
            "with Kosmorrolib v{kosmorrolib_version}"
        ).format(python_version=python_version, kosmorrolib_version=KOSMORROLIB_VERSION)
    )

    return True


def get_args(output_formats: [str]):
    today = date.today()

    parser = argparse.ArgumentParser(
        description=_(
            "Compute the ephemerides and the events for a given date and a given position on Earth."
        ),
        epilog=_(
            "By default, only the events will be computed for today.\n"
            "To compute also the ephemerides, latitude and longitude arguments are needed."
        ),
    )

    parser.add_argument(
        "--version",
        "-v",
        dest="special_action",
        action="store_const",
        const=output_version,
        default=None,
        help=_("Show the program version"),
    )
    parser.add_argument(
        "--format",
        "-f",
        type=str,
        default=None,
        choices=output_formats,
        help=_(
            "The output format. If not provided, it will be inferred from the file extension of the output file."
        ),
    )
    parser.add_argument(
        "--position",
        "-p",
        type=str,
        default=None,
        help=_(
            'The observer\'s position on Earth, in the "latitude,longitude" format. '
            "Can also be set in the KOSMORRO_POSITION environment variable."
        ),
    )
    parser.add_argument(
        "--date",
        "-d",
        type=str,
        default=today.strftime("%Y-%m-%d"),
        help=_(
            "The date for which the ephemerides must be calculated. Can be in the YYYY-MM-DD format "
            'or an interval in the "[+-]YyMmDd" format (with Y, M, and D numbers). '
            "Defaults to current date."
        ).format(default_date=today.strftime("%Y-%m-%d")),
    )
    parser.add_argument(
        "--timezone",
        "-t",
        type=str,
        default=None,
        help=_(
            "The timezone to use to display the hours. It can be either a number (e.g. 1 for UTC+1) or a timezone name (e.g. Europe/Paris). "
            "See https://en.wikipedia.org/wiki/List_of_tz_database_time_zones to find your timezone. "
            "Can also be set in the TZ environment variable."
        ),
    )
    parser.add_argument(
        "--no-colors",
        dest="colors",
        action="store_false",
        help=_("Disable the colors in the console."),
    )
    parser.add_argument(
        "--output",
        "-o",
        type=str,
        default=None,
        help=_("Use this option to save the result in a file."),
    )
    parser.add_argument(
        "--no-graph",
        dest="show_graph",
        action="store_false",
        help=_(
            "Do not generate a graph to represent the rise and set times in the LaTeX or PDF file."
        ),
    )
    parser.add_argument(
        "--debug",
        dest="show_debug_messages",
        action="store_true",
        help=_("Show debugging messages"),
    )

    argcomplete.autocomplete(parser)

    parser.add_argument(
        "--completion",
        type=str,
        help=_("Print a script allowing completion for your shell"),
    )

    return parser.parse_args()


def output_completion(shell: str) -> bool:
    shellcode = argcomplete.shellcode([sys.argv[0]], shell=shell)
    if shellcode == "":
        print_stderr(
            colored(_("No completion script available for this shell."), "red")
        )
        return False

    print(shellcode)
    return True


def main():
    sys.exit(run())



#!/usr/bin/env python3

#    Kosmorro - Compute The Next Ephemerides
#    Copyright (C) 2019  Jérôme Deuchnord <jerome@deuchnord.fr>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
import datetime
import json
import os
import tempfile
import subprocess
import shutil
from pathlib import Path

from babel.dates import format_date, format_time
from tabulate import tabulate
from termcolor import colored

from kosmorrolib import AsterEphemerides, Event, EventType
from kosmorrolib.model import ASTERS, MoonPhase

from .i18n.utils import _
from .i18n import strings
from .exceptions import (
    CompileError,
    UnavailableFeatureError as KosmorroUnavailableFeatureError,
)
from .debug import debug_print
from .utils import KOSMORRO_VERSION


class Dumper(ABC):
    ephemerides: list[AsterEphemerides]
    moon_phase: MoonPhase
    events: list[Event]
    date: datetime.date
    timezone: int
    with_colors: bool
    show_graph: bool

    def __init__(
        self,
        ephemerides: list[AsterEphemerides],
        moon_phase: MoonPhase,
        events: list[Event],
        date: datetime.date,
        timezone: int,
        with_colors: bool,
        show_graph: bool,
    ):
        self.ephemerides = ephemerides
        self.moon_phase = moon_phase
        self.events = events
        self.date = date
        self.timezone = timezone
        self.with_colors = with_colors
        self.show_graph = show_graph

    def get_date_as_string(self, capitalized: bool = False) -> str:
        date = format_date(self.date, "full")

        if capitalized:
            return "".join([date[0].upper(), date[1:]])

        return date

    def __str__(self):
        return self.to_string()

    @abstractmethod
    def to_string(self):
        pass

    @staticmethod
    def is_file_output_needed() -> bool:
        return False


class JsonDumper(Dumper):
    SUPPORTED_EVENTS = [
        EventType.OPPOSITION,
        EventType.CONJUNCTION,
        EventType.OCCULTATION,
        EventType.MAXIMAL_ELONGATION,
        EventType.PERIGEE,
        EventType.APOGEE,
    ]

    def to_string(self):
        return json.dumps(
            {
                "ephemerides": [
                    ephemeris.serialize() for ephemeris in self.ephemerides
                ],
                "moon_phase": self.moon_phase.serialize(),
                "events": list(self.get_events()),
            },
            indent=4,
        )

    def get_events(self) -> [{str: any}]:
        for event in self.events:
            if event.event_type not in self.SUPPORTED_EVENTS:
                continue

            yield event.serialize()


class TextDumper(Dumper):
    def to_string(self):
        text = [self.style(self.get_date_as_string(capitalized=True), "h1")]

        if len(self.ephemerides) > 0:
            text.append(self.stringify_ephemerides())

        text.append(self.get_moon(self.moon_phase))

        if len(self.events) > 0:
            events = self.get_events(self.events)
            if events.strip("\n") != "":
                text.append(
                    "\n".join(
                        [
                            self.style(_("Expected events:"), "h2"),
                            events,
                        ]
                    )
                )

        if self.timezone == 0:
            text.append(self.style(_("Note: All the hours are given in UTC."), "em"))
        else:
            tz_offset = str(self.timezone)
            if self.timezone > 0:
                tz_offset = "".join(["+", tz_offset])
            text.append(
                self.style(
                    _(
                        "Note: All the hours are given in the UTC{offset} timezone."
                    ).format(offset=tz_offset),
                    "em",
                )
            )

        return "\n\n".join(text)

    def style(self, text: str, tag: str) -> str:
        if not self.with_colors:
            return text

        styles = {
            "h1": lambda t: colored(t, "green", attrs=["bold"]),
            "h2": lambda t: colored(t, "magenta", attrs=["bold"]),
            "th": lambda t: colored(t, attrs=["bold"]),
            "strong": lambda t: colored(t, attrs=["bold"]),
            "em": lambda t: colored(t, attrs=["dark"]),
        }

        return styles.get(tag, lambda t: t)(text)

    def stringify_ephemerides(self) -> str:
        data = []

        for ephemeris in self.ephemerides:
            object_name = strings.from_object(ephemeris.object.identifier)
            if object_name is None:
                continue

            name = self.style(object_name, "th")

            planet_rise = (
                "-"
                if ephemeris.rise_time is None
                else format_time(ephemeris.rise_time, "short")
            )
            planet_culmination = (
                "-"
                if ephemeris.culmination_time is None
                else format_time(ephemeris.culmination_time, "short")
            )
            planet_set = (
                "-"
                if ephemeris.set_time is None
                else format_time(ephemeris.set_time, "short")
            )

            data.append([name, planet_rise, planet_culmination, planet_set])

        return tabulate(
            data,
            headers=[
                self.style(_("Object"), "th"),
                self.style(_("Rise time"), "th"),
                self.style(_("Culmination time"), "th"),
                self.style(_("Set time"), "th"),
            ],
            tablefmt="simple",
            stralign="center",
            colalign=("left",),
        )

    def get_events(self, events: [Event]) -> str:
        data = []

        for event in events:
            description = strings.from_event(event)
            if description is None:
                continue

            data.append(
                [
                    self.style(format_time(event.start_time, "short"), "th"),
                    description,
                ]
            )

        return tabulate(data, tablefmt="plain", stralign="left")

    def get_moon(self, moon_phase: MoonPhase) -> str:
        if moon_phase is None:
            return _("Moon phase is unavailable for this date.")

        current_moon_phase = self.style(
            strings.from_moon_phase(moon_phase.phase_type), "strong"
        )
        new_moon_phase = _(
            "{next_moon_phase} on {next_moon_phase_date} at {next_moon_phase_time}"
        ).format(
            next_moon_phase=_(strings.from_moon_phase(moon_phase.get_next_phase())),
            next_moon_phase_date=format_date(moon_phase.next_phase_date, "full"),
            next_moon_phase_time=format_time(moon_phase.next_phase_date, "short"),
        )

        return "\n".join([current_moon_phase, new_moon_phase])


class LatexDumper(Dumper):
    def to_string(self):
        template_path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            "assets",
            "latex",
            "template.tex",
        )

        with open(template_path, mode="r") as file:
            template = file.read()

        return self._make_document(template)

    def _make_document(self, template: str) -> str:
        kosmorro_logo_path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            "assets",
            "png",
            "kosmorro-logo.png",
        )

        moon_phase_graphics = os.path.join(
            os.path.abspath(os.path.dirname(__file__)),
            "assets",
            "moonphases",
            "png",
            ".".join(
                [self.moon_phase.phase_type.name.lower().replace("_", "-"), "png"]
            ),
        )

        document = template

        if len(self.ephemerides) == 0:
            document = self._remove_section(document, "ephemerides")

        if len(self.events) == 0:
            document = self._remove_section(document, "events")

        document = self.add_strings(document, kosmorro_logo_path, moon_phase_graphics)

        if self.show_graph:
            # The graphephemerides environment beginning tag must end with a percent symbol to ensure
            # that no extra space will interfere with the graph.
            document = document.replace(
                r"\begin{ephemerides}", r"\begin{graphephemerides}%"
            ).replace(r"\end{ephemerides}", r"\end{graphephemerides}")

        return document

    def add_strings(
        self, document: str, kosmorro_logo_path: str, moon_phase_graphics: str
    ) -> str:
        document = document.replace(
            "+++CURRENT-DATE+++", datetime.datetime.now().isoformat()
        )
        document = document.replace("+++KOSMORRO-VERSION+++", KOSMORRO_VERSION)
        document = document.replace("+++KOSMORRO-LOGO+++", kosmorro_logo_path)
        document = document.replace("+++DOCUMENT-TITLE+++", _("Overview of your sky"))
        document = document.replace(
            "+++DOCUMENT-DATE+++", self.get_date_as_string(capitalized=True)
        )
        document = document.replace(
            "+++INTRODUCTION+++",
            "\n\n".join(
                [
                    _(
                        "This document summarizes the ephemerides and the events of {date}. "
                        "It aims to help you to prepare your observation session. "
                        "All the hours are given in {timezone}."
                    ).format(
                        date=self.get_date_as_string(),
                        timezone=(
                            "UTC+%d" % self.timezone if self.timezone != 0 else "UTC"
                        ),
                    ),
                    _(
                        "Don't forget to check the weather forecast before you go out with your equipment."
                    ),
                ]
            ),
        )
        document = document.replace(
            "+++SECTION-EPHEMERIDES+++", _("Ephemerides of the day")
        )
        document = document.replace("+++EPHEMERIDES-OBJECT+++", _("Object"))
        document = document.replace("+++EPHEMERIDES-RISE-TIME+++", _("Rise time"))
        document = document.replace(
            "+++EPHEMERIDES-CULMINATION-TIME+++", _("Culmination time")
        )
        document = document.replace("+++EPHEMERIDES-SET-TIME+++", _("Set time"))
        document = document.replace("+++EPHEMERIDES+++", self._make_ephemerides())
        document = document.replace("+++GRAPH_LABEL_HOURS+++", _("hours"))
        document = document.replace("+++MOON-PHASE-GRAPHICS+++", moon_phase_graphics)
        document = document.replace("+++CURRENT-MOON-PHASE-TITLE+++", _("Moon phase:"))
        document = document.replace(
            "+++CURRENT-MOON-PHASE+++",
            strings.from_moon_phase(self.moon_phase.phase_type),
        )
        document = document.replace("+++SECTION-EVENTS+++", _("Expected events"))
        document = document.replace("+++EVENTS+++", self._make_events())

        for aster in ASTERS:
            object_name = strings.from_object(aster.identifier)
            if object_name is None:
                continue

            document = document.replace(
                "+++ASTER_%s+++" % aster.identifier.name,
                object_name,
            )

        return document

    def _make_ephemerides(self) -> str:
        latex = []
        graph_y_component = 18

        if self.ephemerides is not None:
            for ephemeris in self.ephemerides:
                aster_rise = (
                    "-"
                    if ephemeris.rise_time is None
                    else format_time(ephemeris.rise_time, "short")
                )
                aster_culmination = (
                    "-"
                    if ephemeris.culmination_time is None
                    else format_time(ephemeris.culmination_time, "short")
                )
                aster_set = (
                    "-"
                    if ephemeris.set_time is None
                    else format_time(ephemeris.set_time, "short")
                )

                if not self.show_graph:
                    object_name = strings.from_object(ephemeris.object.identifier)
                    if object_name is not None:
                        latex.append(
                            r"\object{%s}{%s}{%s}{%s}"
                            % (
                                object_name,
                                aster_rise,
                                aster_culmination,
                                aster_set,
                            )
                        )
                else:
                    if ephemeris.rise_time is not None:
                        raise_hour = ephemeris.rise_time.hour
                        raise_minute = ephemeris.rise_time.minute
                    else:
                        raise_hour = raise_minute = 0
                        aster_rise = ""

                    if ephemeris.set_time is not None:
                        set_hour = ephemeris.set_time.hour
                        set_minute = ephemeris.set_time.minute
                    else:
                        set_hour = 24
                        set_minute = 0
                        aster_set = ""
                    sets_after_end = set_hour > raise_hour

                    if not sets_after_end:
                        latex.append(
                            r"\graphobject{%d}{gray}{0}{0}{%d}{%d}{}{%s}"
                            % (graph_y_component, set_hour, set_minute, aster_set)
                        )
                        set_hour = 24
                        set_minute = 0

                    latex.append(
                        r"\graphobject{%d}{gray}{%d}{%d}{%d}{%d}{%s}{%s}"
                        % (
                            graph_y_component,
                            raise_hour,
                            raise_minute,
                            set_hour,
                            set_minute,
                            aster_rise,
                            aster_set if sets_after_end else "",
                        )
                    )
                    graph_y_component -= 2

        return "".join(latex)

    def _make_events(self) -> str:
        latex = []

        for event in self.events:
            event_name = strings.from_event(event)
            if event_name is None:
                continue

            latex.append(
                r"\event{%s}{%s}" % (format_time(event.start_time, "short"), event_name)
            )

        return "".join(latex)

    @staticmethod
    def _remove_section(document: str, section: str):
        begin_section_tag = "%%%%%% BEGIN-%s-SECTION" % section.upper()
        end_section_tag = "%%%%%% END-%s-SECTION" % section.upper()

        document = document.split("\n")
        new_document = []

        ignore_line = False
        for line in document:
            if begin_section_tag in line or end_section_tag in line:
                ignore_line = not ignore_line
                continue
            if ignore_line:
                continue
            new_document.append(line)

        return "\n".join(new_document)


class PdfDumper(Dumper):
    def to_string(self):
        try:
            latex_dumper = LatexDumper(
                self.ephemerides,
                self.moon_phase,
                self.events,
                date=self.date,
                timezone=self.timezone,
                with_colors=self.with_colors,
                show_graph=self.show_graph,
            )

            return self._compile(latex_dumper.to_string())
        except RuntimeError as error:
            raise KosmorroUnavailableFeatureError(
                _(
                    "Building PDF was not possible, because some dependencies are not"
                    " installed.\nPlease look at the documentation at https://kosmorro.space/cli/generate-pdf/ "
                    "for more information."
                )
            ) from error

    @staticmethod
    def is_file_output_needed() -> bool:
        return True

    @staticmethod
    def _compile(latex_input) -> bytes:
        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        current_dir = (
            os.getcwd()
        )  # Keep the current directory to return to it after the PDFLaTeX execution

        try:
            temp_dir = tempfile.mkdtemp()
            temp_tex = "%s/%s.tex" % (temp_dir, timestamp)

            with open(temp_tex, "w") as tex_file:
                tex_file.write(latex_input)

            os.chdir(temp_dir)
            debug_print("LaTeX content:\n%s" % latex_input)

            subprocess.run(
                ["pdflatex", "-interaction", "nonstopmode", "%s.tex" % timestamp],
                capture_output=True,
                check=True,
            )

            os.chdir(current_dir)

            with open("%s/%s.pdf" % (temp_dir, timestamp), "rb") as pdffile:
                return bytes(pdffile.read())

        except FileNotFoundError as error:
            raise KosmorroUnavailableFeatureError(
                "TeXLive is not installed."
            ) from error

        except subprocess.CalledProcessError as error:
            with open("/tmp/kosmorro-%s.log" % timestamp, "wb") as file:
                file.write(error.stdout)

            raise CompileError(
                _(
                    "An error occurred during the compilation of the PDF.\n"
                    "Please open an issue at https://github.com/Kosmorro/kosmorro/issues and share "
                    "the content of the log file at /tmp/kosmorro-%s.log" % timestamp
                )
            ) from error



#!/usr/bin/env python3
from datetime import datetime, date

import pytz
from termcolor import colored as do_color
from sys import stderr

from importlib.metadata import version

KOSMORRO_VERSION = version("kosmorro")
KOSMORROLIB_VERSION = version("kosmorrolib")


global _COLORS_ACTIVATED


def set_colors_activated(activated: bool):
    global _COLORS_ACTIVATED
    _COLORS_ACTIVATED = activated


def colored(text, color=None, on_color=None, attrs=None):
    """Decorator to use colors only when they are activated"""
    if not _COLORS_ACTIVATED:
        return text

    return do_color(text, color, on_color, attrs)


def print_stderr(*values: object):
    print(*values, file=stderr)


def get_timezone(value: int | str, dt: datetime | date) -> float:
    try:
        timezone = float(value)
    except ValueError:
        if isinstance(dt, date):
            dt = datetime.combine(dt, datetime.min.time())
        timezone = pytz.timezone(value).utcoffset(dt).total_seconds() / 3600

    return timezone



#!/usr/bin/env python3

from typing import Union
from babel.dates import format_time
from .utils import _
from kosmorrolib import (
    EventType,
    MoonPhaseType,
    ObjectIdentifier,
    Event,
    SeasonType,
    LunarEclipseType,
)


def from_event(event: Event) -> Union[None, str]:
    string = None
    match event.event_type:
        case EventType.OPPOSITION:
            string, details = (
                _("%s is in opposition") % from_object(event.objects[0].identifier),
                None,
            )
        case EventType.CONJUNCTION:
            string, details = (
                _("%s and %s are in conjunction")
                % (
                    from_object(event.objects[0].identifier),
                    from_object(event.objects[1].identifier),
                ),
                None,
            )
        case EventType.OCCULTATION:
            string, details = (
                _("%s occults %s")
                % (
                    from_object(event.objects[0].identifier),
                    from_object(event.objects[1].identifier),
                ),
                None,
            )
        case EventType.MAXIMAL_ELONGATION:
            string, details = (
                _("Elongation of %s is maximal")
                % from_object(event.objects[0].identifier),
                lambda e: "{:.3n}°".format(e.details["deg"]),
            )
        case EventType.PERIGEE:
            string, details = (
                _("%s is at its periapsis") % from_object(event.objects[0].identifier),
                None,
            )
        case EventType.APOGEE:
            string, details = (
                _("%s is at its apoapsis") % from_object(event.objects[0].identifier),
                None,
            )
        case EventType.SEASON_CHANGE:
            match event.details["season"]:
                case SeasonType.MARCH_EQUINOX:
                    string = _("March equinox")
                case SeasonType.JUNE_SOLSTICE:
                    string = _("June solstice")
                case SeasonType.SEPTEMBER_EQUINOX:
                    string = _("September equinox")
                case _:
                    string = _("December solstice")

            details = None
        case EventType.LUNAR_ECLIPSE:
            match event.details["type"]:
                case LunarEclipseType.TOTAL:
                    string = _("Total lunar eclipse until %(hour)s") % {
                        "hour": format_time(event.end_time, "short")
                    }

                case LunarEclipseType.PENUMBRAL:
                    string = _("Penumbral lunar eclipse until %(hour)s") % {
                        "hour": format_time(event.end_time, "short")
                    }

                case LunarEclipseType.PARTIAL:
                    string = _("Partial lunar eclipse until %(hour)s") % {
                        "hour": format_time(event.end_time, "short")
                    }

            details = None
        case _:
            return None

    if details is not None:
        return "%s (%s)" % (string, details(event))

    return string


def from_moon_phase(moon_phase: MoonPhaseType) -> str:
    string = {
        MoonPhaseType.NEW_MOON: _("New Moon"),
        MoonPhaseType.WAXING_CRESCENT: _("Waxing Crescent"),
        MoonPhaseType.FIRST_QUARTER: _("First Quarter"),
        MoonPhaseType.WAXING_GIBBOUS: _("Waxing Gibbous"),
        MoonPhaseType.FULL_MOON: _("Full Moon"),
        MoonPhaseType.WANING_GIBBOUS: _("Waning Gibbous"),
        MoonPhaseType.LAST_QUARTER: _("Last Quarter"),
        MoonPhaseType.WANING_CRESCENT: _("Waning Crescent"),
    }.get(moon_phase)

    if string is None:
        raise RuntimeError("Unknown moon phase: %s." % moon_phase)

    return string


def from_object(identifier: ObjectIdentifier) -> str:
    return {
        ObjectIdentifier.SUN: _("Sun"),
        ObjectIdentifier.MOON: _("Moon"),
        ObjectIdentifier.MERCURY: _("Mercury"),
        ObjectIdentifier.VENUS: _("Venus"),
        ObjectIdentifier.EARTH: _("Earth"),
        ObjectIdentifier.MARS: _("Mars"),
        ObjectIdentifier.JUPITER: _("Jupiter"),
        ObjectIdentifier.SATURN: _("Saturn"),
        ObjectIdentifier.URANUS: _("Uranus"),
        ObjectIdentifier.NEPTUNE: _("Neptune"),
        ObjectIdentifier.PLUTO: _("Pluto"),
    }.get(identifier)



#!/usr/bin/env python3

import re

from kosmorrolib import Position

from .i18n.utils import _


def get_position(from_str: str) -> Position:
    if not re.search(r"^([\d.-]+)[,;]([\d.-]+)$", from_str):
        raise ValueError(_("The given position (%s) is not valid." % from_str))

    latitude_longitude = from_str.split(";")
    if len(latitude_longitude) == 1:
        latitude_longitude = from_str.split(",")

    return Position(float(latitude_longitude[0]), float(latitude_longitude[1]))



#!/usr/bin/env python3

from traceback import print_exception

show_debug_messages = False


def debug_print(what, force: bool = False):
    if not force and not show_debug_messages:
        return

    if isinstance(what, Exception):
        print_exception(type(what), value=what, tb=None)
    else:
        print("[DEBUG] %s" % what)



#!/usr/bin/env python3

import re

from dateutil.relativedelta import relativedelta
from datetime import date

from kosmorro.i18n.utils import _


def parse_date(date_arg: str) -> date:
    if re.match(r"^\d{4}-\d{2}-\d{2}$", date_arg):
        try:
            return date.fromisoformat(date_arg)
        except ValueError as error:
            raise ValueError(
                _("The date {date} is not valid: {error}").format(
                    date=date_arg, error=error.args[0]
                )
            ) from error
    elif re.match(r"^([+-])(([0-9]+)y)?[ ]?(([0-9]+)m)?[ ]?(([0-9]+)d)?$", date_arg):

        def get_offset(date_arg: str, signifier: str):
            if re.search(r"([0-9]+)" + signifier, date_arg):
                return abs(
                    int(re.search(r"[+-]?([0-9]+)" + signifier, date_arg).group(0)[:-1])
                )
            return 0

        days = get_offset(date_arg, "d")
        months = get_offset(date_arg, "m")
        years = get_offset(date_arg, "y")

        if date_arg[0] == "+":
            return date.today() + relativedelta(days=days, months=months, years=years)
        return date.today() - relativedelta(days=days, months=months, years=years)

    else:
        error_msg = _(
            "The date {date} does not match the required YYYY-MM-DD format or the offset format."
        )
        raise ValueError(error_msg.format(date=date_arg))



#!/usr/bin/env python3

#    Kosmorro - Compute The Next Ephemerides
#    Copyright (C) 2019  Jérôme Deuchnord <jerome@deuchnord.fr>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from datetime import date
from babel.dates import format_date
from kosmorro.i18n.utils import _


class UnavailableFeatureError(RuntimeError):
    def __init__(self, msg: str):
        super().__init__()
        self.msg = msg


class OutOfRangeDateError(RuntimeError):
    def __init__(self, min_date: date, max_date: date):
        super().__init__()
        self.min_date = min_date
        self.max_date = max_date
        self.msg = _(
            "The date must be between {minimum_date} and {maximum_date}"
        ).format(
            minimum_date=format_date(min_date, "long"),
            maximum_date=format_date(max_date, "long"),
        )


class InvalidOutputFormatError(RuntimeError):
    def __init__(self, output_format: str, accepted_extensions: [str]):
        super().__init__()
        self.output_format = output_format
        self.accepted_extensions = accepted_extensions
        self.msg = _(
            "Invalid output format: {output_format}. Output file must end with: {accepted_extensions}"
        ).format(
            output_format=output_format,
            accepted_extensions=", ".join(accepted_extensions),
        )


class CompileError(RuntimeError):
    def __init__(self, msg):
        super().__init__()
        self.msg = msg



#!/usr/bin/env python3

#    Kosmorro - Compute The Next Ephemerides
#    Copyright (C) 2019  Jérôme Deuchnord <jerome@deuchnord.fr>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import re
from pathlib import Path

CACHE_FOLDER = str(Path.home()) + "/.kosmorro-cache"
NO_COLOR = "NO_COLOR" in os.environ


class Environment:
    def __init__(self):
        self._vars = {}

    def __set__(self, key, value):
        self._vars[key] = value

    def __getattr__(self, key):
        return self._vars[key] if key in self._vars else None

    def __str__(self):
        return self._vars.__str__()

    def __len__(self):
        return len(self._vars)


def get_env_vars() -> Environment:
    environment = Environment()

    environment.tz = os.getenv("TZ")

    for var in os.environ:
        if not re.search("^KOSMORRO_", var):
            continue

        [_, env] = var.split("_", 1)
        environment.__set__(env.lower(), os.getenv(var))

    return environment
